# Fontant
import matplotlib
from matplotlib.font_manager import findSystemFonts
from matplotlib import ft2font
import os

def get_fonts():
    # Get default matplotlib paths
    paths = [os.path.join(matplotlib.get_data_path(), 'fonts', subdir)
             for subdir in ['ttf', 'afm', 'pdfcorefonts']]
    # Get paths from OS
    for pathname in ['TTFPATH', 'AFMPATH']:
        if pathname in os.environ:
            ttfpath = os.environ[pathname]
            if ttfpath.find(';') >= 0:  # win32 style
                paths.extend(ttfpath.split(';'))
            elif ttfpath.find(':') >= 0:  # unix style
                paths.extend(ttfpath.split(':'))
            else:
                paths.append(ttfpath)
    # "ttf" here means "otf" and "ttc" too
    fonts_ttf_mpl = findSystemFonts(paths, fontext="ttf")
    fonts_ttf_sys = findSystemFonts(fontext="ttf")
    # We are going to try to avoid postscript fonts at all costs
    # fonts_afm_mpl = font_manager.findSystemFonts(paths, fontext="afm")
    # fonts_afm_sys = font_manager.findSystemFonts(fontext="afm")
    return fonts_ttf_mpl + fonts_ttf_sys# + fonts_afm_mpl + fonts_afm_sys

WEIGHTS = {
    # Originally from fontconfig's FcFreeTypeQueryFaceInternal, here backported
    # from latest matplotlib
    "thin": 100,
    "extralight": 200,
    "extra light": 200,
    "ultralight": 200,
    "ultra light": 200,
    "demilight": 350,
    "demi light": 350,
    "semilight": 350,
    "semi light": 350,
    "light": 300,  # Needs to come *after* demi/semilight!
    "book": 380,
    "regular": 400,
    "roman": 400,
    "normal": 400,
    "medium": 500,
    "demibold": 600,
    "demi": 600,
    "semibold": 600,
    "extrabold": 800,
    "extra bold": 800,
    "superbold": 800,
    "super bold": 800,
    "ultrabold": 800,
    "ultra bold": 800,
    "bold": 700,  # Needs to come *after* extra/super/ultrabold!
    "ultrablack": 1000,
    "ultra black": 1000,
    "superblack": 1000,
    "super black": 1000,
    "extrablack": 1000,
    "extra black": 1000,
    "ultra": 1000,
    "black": 900,  # Needs to come *after* ultra/super/extrablack!
    "heavy": 900,
}

STRETCH = {
    'ultracondensed':  -1,
    'ultra-condensed': 1,
    'extracondensed':  -1,
    'extra-condensed': 2,
    'narrow':          -1,
    'condensed':       3,
    'semicondensed':   -1,
    'semi-condensed':  4,
    'normal':          5,
    'semiexpanded':    -1,
    'semi-expanded':   6,
    'semiextended':    -1,
    'semi-extended':   -1,
    'expanded':        7,
    'extended':        -1,
    'extraexpanded':   -1,
    'extra-expanded':  8,
    'extraextended':   -1,
    'extra-extended':  -1,
    'ultraexpanded':   -1,
    'ultra-expanded':  9,
    'ultraextended':   -1,
    'ultra-extended':  -1,
}

SPECIALS = [
    'sc',
    'caps',
    'alt',
]

FOUNDRIES = {
    "Adobe": "Adobe",
    "Bigelow": "B&H",
    "Bitstream": "Bitstream",
    "Gnat": "Culmus",
    "Iorsh": "Culmus",
    "HanYang System": "Hanyang",
    "Font21": "Hwan",
    "Google": "Google",
    "IBM": "IBM",
    "International Typeface Corporation": "ITC",
    "ITC": "ITC",
    "Linotype": "Linotype",
    "LINOTYPE-HELL": "Linotype",
    "Microsoft": "Microsoft",
    "Monotype": "Monotype",
    "Omega": "Omega",
    "STIX": "STIX",
    "TeX": "TeX",
    "Tiro Typeworks": "Tiro",
    "URW": "URW",
    "XFree86": "XFree86",
    "Xorg": "Xorg",
}

FOUNDRIES_NAME = {
    "Adobe": "Adobe",
    "A2": "A2",
    "ADF": "Arkandis",
    "AEF": "Altered Ego Fonts",
    "AS": "Alphabet Soup",
    "AT": "Agfa",
    "BT": "Bitstream",
    "CC": "Comicraft",
    "CG": "Compugraphic",
    "EF": "Elsner+Flake",
    "FB": "Font Bureau",
    "F2F": "Face2Face",
    "FF": "FontFont",
    "FP": "FontPartners",
    "FTN": "Fountain",
    "GFS": "Greek Font Society",
    "ITC": "ITC",
    "LFT": "Leftloft",
    "LP": "Letter Perfect",
    "LT": "Linotype",
    "LTC": "Lanston",
    "MT": "Monotype",
    "MVB": "MVB",
    "ND": "Neufville Digital",
    "P22": "P22",
    "PL": "Photo Lettering",
    "PT": "Paratype",
    "PTF": "Porchez",
    "PTL": "Primetype",
    "RB": "Richard Beatty",
    "RTF": "Rimmer",
    "SG": "Scangraphic",
    "TC": "Typesettra",
    "URW": "URW++",
    "URW++": "URW++",
    "WTC": "WTC",
}

OPTICAL_SIZES = ["display", "caption", "subhead", "re",
                 "text", "headline", "banner", "micro", "deck"]

def get_property(font, props):
    PROPS = {
        "ps_notice": ("ps", 1), # Copyright notice
        "ps_fullname": ("ps", 2),
        "ps_familyname": ("ps", 3),
        "ps_weight": ("ps", 4),
        "ps_italicangle": ("ps", 5),
        "ps_monospace": ("ps", 6),
        "os2_vendorid": ("os2", "achVendID"),
        "os2_widthclass": ("os2", "usWidthClass"),
        "os2_weightclass": ("os2", "usWeightClass"),
        #"os2_familytype": ("panose", 0), # panose is a weird string from the OS/2 table
        #"os2_weight": ("panose", 2),
        "sfnt_copyright": ("sfnt", 0),
        "sfnt_family": ("sfnt", 1), # Only four fonts can share this: normal, italic/oblique, bold, and bold+italic/oblique.  For fonts with more variants, use 16 and 17.
        "sfnt_subfamily": ("sfnt", 2), # For style and weight variants only
        "sfnt_identifier": ("sfnt", 3), # Unique font identifier
        "sfnt_fullname": ("sfnt", 4), # Combines 1/2/16/17, but only when each of these is relevant (e.g. "regular" is not relevant for Arial Black)
        "sfnt_version": ("sfnt", 5), # Version string
        "sfnt_postscriptname": ("sfnt", 6),
        "sfnt_trademark": ("sfnt", 7),
        "sfnt_manufacturer": ("sfnt", 8),
        "sfnt_designer": ("sfnt", 9),
        "sfnt_description": ("sfnt", 10),
        "sfnt_vendorurl": ("sfnt", 11),
        "sfnt_designerurl": ("sfnt", 12),
        "sfnt_licensedescription": ("sfnt", 13), # In plain English, describe how the font can be used
        "sfnt_licenseurl": ("sfnt", 14),
        "sfnt_typographicfamily": ("sfnt", 16), # Just the overall font name with no subtypes included, e.g. "Helvetica Neue LT"
        "sfnt_typographicsubfamily": ("sfnt", 17), # All subtypes, e.g. "Semibold Caption Thin"
        "sfnt_wwsfamily": ("sfnt", 21), # Base name and everything except weight and italics
        "sfnt_wwssubfamily": ("sfnt", 22), # Weight and italic style
    }
    # For example, Minion Pro Semibold Italic Caption:
    # sfnt_familyname:           Minion Pro SmBd Capt
    # sfnt_subfamilyname:        Italic
    # sfnt_typographicfamily:    Minion Pro
    # sfnt_typographicsubfamily: Semibold Italic Caption
    # sfnt_wwsfamily:            Minion Pro Caption
    # sfnt_wwssubfamily:         Semibold Italic
    # For more info see https://docs.microsoft.com/en-us/typography/opentype/spec/name#name-ids
    # For more info on postscript fields, see:
    # https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6name.html
    # https://docs.microsoft.com/en-us/typography/opentype/spec/post


    returns = []
    returnstr = False
    if isinstance(props, str):
        props = [props]
        returnstr = True
    for prop in props:
        table, key = PROPS[prop]
        if table == "ps":
            try:
                psinfo = font.get_ps_font_info()
                returns.append(psinfo[key])
            except:
                pass
        elif table == "os2":
            os2 = font.get_sfnt_table("OS/2")
            if os2 and os2['version'] != 0xffff:
                if os2.get(key, ""):
                    returns.append(os2[key])
        elif table == "sfnt":
            sfnt_name = font.get_sfnt()
            keyval = sfnt_name.get((3,1,1033,key), b"").decode('utf_16_be')
            if keyval:
                returns.append(keyval)
            else:
                keyval = sfnt_name.get((1,0,0,key), b"").decode('latin-1')
                if keyval:
                    returns.append(keyval)
        else:
            raise ValueError("Invalid property passed: "+prop)
    if returnstr:
        return returns[0] if len(returns) == 1 else None
    return returns

def get_familyname(font):
    names = get_property(font, ["sfnt_typographicfamily", "ps_familyname",
                                "sfnt_family", "sfnt_wwsfamily", "ps_fullname",
                                "sfnt_fullname"])
    return (names+[""])[0]

def get_fullname(font):
    names = get_property(font, ["ps_fullname", "sfnt_fullname"])
    return (names+[""])[0]

def get_style(font):
    # First check the fields devoted to this.  Even though the sfnt
    # subfamily is supposed to have the best version of this information, we
    # check the postscript full font name first because sometimes the others
    # mix up italic and oblique.  If "oblique" is in the full name, we know for
    # sure it shouldn't be italic.
    familynames = get_property(font, ["ps_fullname", "sfnt_typographicsubfamily",
                                      "sfnt_wwssubfamily", "sfnt_subfamily",
                                      "sfnt_fullname"])
    for familyname in familynames:
        for style in ["italic", "oblique", "kursiv"]:
            if style in familyname.lower().replace("-", " ").split(" "):
                return style
    # Now check properties of the font
    if font.style_flags & ft2font.ITALIC:
        return "italic"
    # if get_property(font, "ps_italicangle") != 0:
    #     return "italic"
    return "normal"

def get_monospace(font):
    # First check if it is in postscript information.  If so, this makes our
    # life easy.
    ismono = get_property(font, "ps_monospace")
    if ismono is not None:
        return bool(ismono)
    # If not, use the font object
    return bool(font.face_flags & ft2font.FIXED_WIDTH)
    # # If not, let's render two characters that should be different widths.  If
    # # they are the same width, we can assume the font is monospace.
    # font.set_text(".")
    # width_dot = font.get_width_height()[0]
    # font.set_text("M")
    # width_M = font.get_width_height()[0]
    # return (width_dot == width_M)


def get_foundry(font):
    # First try to extract it from the name
    names = get_property(font, ["sfnt_typographicfamily", "sfnt_family",
                                "ps_familyname", "ps_fullname", "sfnt_fullname"])
    for name in names:
        for abbrev,fname in FOUNDRIES_NAME.items():
            if abbrev in name.replace("-", " ").split(" "):
                return fname
    # Then extract from "manufacturer" text
    manufac = get_property(font, "sfnt_manufacturer")
    if manufac:
        for ftext,fname in FOUNDRIES.items():
            if ftext in manufac:
                return fname
        return manufac
    # Now look in license text
    licensetexts = get_property(font, ["sfnt_trademark", "ps_notice",
                                       "sfnt_copyright"])
    for licensetext in licensetexts:
        for ftext,fname in FOUNDRIES.items():
            if ftext in licensetext:
                return fname
        for fname in FOUNDRIES_NAME.values():
            if fname in licensetext:
                return fname
    return ""

def get_opticalsize(font):
    # The only way I can find to get this is from the name.
    names = get_property(font, ["sfnt_typographicsubfamily",
                                "sfnt_wwsfamily", "sfnt_family",
                                "ps_fullname", "sfnt_fullname"])
    for name in names:
        for opsz in OPTICAL_SIZES:
            if opsz in name.lower().replace("-", " ").split(" "):
                return opsz
    return "normal"

def get_stretch(font):
    # First try based on the name
    names = get_property(font, ["sfnt_typographicsubfamily",
                                "sfnt_wwsfamily", "sfnt_family",
                                "ps_fullname", "sfnt_fullname"])
    for name in names:
        for stretch in STRETCH.keys():
            if stretch in name.lower().replace("-", " ").split(" "):
                return stretch
    # If the name failed, then try using os2 information
    width = get_property(font, "os2_weightclass")
    if width:
       for stretch,val in STRETCH.items():
           if val == width:
               return stretch
    return "normal"

def get_weightnumber(font):
    # First try the OS/2 table weight class
    os2weight = get_property(font, "os2_weightclass")
    if os2weight:
        return os2weight
    # Otherwise, try to get it from the name
    weightname = get_weight(font)
    return WEIGHTS.get(weightname, None)

def get_weight(font):
    # First, try to extract it from the font name
    weightnames = get_property(font, ["sfnt_typographicsubfamily",
                                      "sfnt_subfamily", "sfnt_wwssubfamily",
                                      "ps_fullname", "sfnt_fullname"])
    for weightname in weightnames:
        for w in WEIGHTS.keys():
            if w in weightname.lower().replace("-", " ").split(" ") or \
               (" " in w and w in weightname.lower()): # Some weight names have spaces
                return w
    # Now try postscript weight.  We don't do this first because this field is
    # often wrong.
    weight = get_property(font, "ps_weight")
    if weight:
        return weight.lower()
    # As a last ditch, look at flags
    if font.style_flags & ft2font.BOLD:
        return "bold"
    return "regular"

def get_identifier(font):
    # First, get the item guaranteed to be a unique identifier
    uniqueid = get_property(font, "sfnt_identifier")
    if uniqueid:
        return uniqueid
    # If that doesn't work, just use the filename and call it a day.
    return os.path.split(font.fname)[-1]

def get_special(font):
    varnames = get_property(font, ["sfnt_typographicsubfamily",
                                   "sfnt_subfamily", "sfnt_wwssubfamily",
                                   "ps_fullname", "sfnt_fullname"])
    allspecs = []
    for varname in varnames:
        for s in SPECIALS:
            if s in varname.lower().replace("-", " ").split(" "):
                allspecs.append(s)
    if allspecs:
        return " ".join(sorted(set(allspecs)))
    else:
        return "none"

def get_version(font):
    return get_property(font, "sfnt_version")

def get_base_style(font):
    base_style = font.style_flags
    # Sometimes italic/bold versions are mistakenly detected as base_style 0.
    if base_style == 0:
        if get_weight(font) == "bold":
            base_style += 2
        if get_style(font) != "normal":
            base_style += 1
    return base_style

def loadttf(path):
    try:
        font = ft2font.FT2Font(path)
    except RuntimeError:
        return None
    props = {
        "fname": font.fname,
        "full_name": get_fullname(font),
        "postscript_name": font.postscript_name,
        "family_name": get_familyname(font),
        "style": get_style(font),
        "weight": get_weight(font),
        "weight_number": get_weightnumber(font),
        "foundry": get_foundry(font),
        "stretch": get_stretch(font),
        "opticalsize": get_opticalsize(font),
        "monospace": get_monospace(font),
        "identifier": get_identifier(font),
        "version": get_version(font),
        "special": get_special(font),
        "num_faces": font.num_faces,
        "base_style": get_base_style(font), # 0 = regular, 1 = italic, 2 = bold, 3 = bolditalic
    }
    return props

class NoFontFoundError(ValueError):
    pass

class MultipleFontsFoundError(ValueError):
    pass

_find_font_cache = {}

def find_font(name, *, weight=None, style=None, stretch=None, opticalsize=None, monospace=None, foundry=None, special=None, multiple=False):
    global _find_font_cache
    cachename = (name, weight, style, stretch, opticalsize, monospace, foundry, special, multiple)
    if cachename in _find_font_cache.keys():
        return _find_font_cache[cachename]
    # Check if they passed a filename for the name.  If so, return that.
    if name is not None and os.path.isfile(name):
        loadedttf = loadttf(name)
        if loadedttf is None:
            print(f"Warning: font path {name} could not be loaded, trying a more thorough search")
        else:
            _find_font_cache[cachename] = loadedttf
            return loadedttf
    # If not, firs, find all fonts which contain the specified name as a
    # substring.  This first expression uses the "for lf in [...]" hack like a
    # "let" statement in lisp so that we can still use list comprehensions and
    # filter for None without running the loadttf function twice.  (I think
    # there might be special syntax for this in newer versions of Python?)
    fonts = [lf for p in get_fonts() for lf in [loadttf(p)] if lf is not None]
    fonts = [f for f in fonts if name.lower() in f["full_name"].lower() or
                                 name.lower() in f["family_name"].lower() or
                                 name.lower() in f["postscript_name"].lower()]
    if len(fonts) == 0:
        raise NoFontFoundError("Invalid font name")
    # Next, try to narrow it down based on family name.
    _fonts = [f for f in fonts if name.lower() in f["family_name"].lower()]
    fonts = _fonts or fonts
    # Next, check if there is an exact match in full name, family name, or postscript name
    _fonts = [f for f in fonts if name.lower() in [f["family_name"].lower(),
                                                   f["full_name"].lower(),
                                                   f["postscript_name"].lower()]]
    fonts = _fonts or fonts
    # Now we filter by each of the properties which was passed as an argument
    def filter_if_exists(fonts, prop, val):
        lower = lambda x : x.lower() if isinstance(x, str) else x
        if val is None:
            return fonts
        allprops = set(lower(f[prop]) for f in fonts)
        if lower(val) in allprops:
            return [f for f in fonts if lower(val) == lower(f[prop])]
        else:
            errortext = f"No selected fonts have {prop} = \"{val}\", options are:\n    "
            errortext += ", ".join(f'"{p}"' for p in sorted(allprops))
            raise NoFontFoundError(errortext)
    if isinstance(weight, int):
        fonts = filter_if_exists(fonts, "weight_number", weight)
    elif isinstance(weight, str):
        fonts = filter_if_exists(fonts, "weight", weight)
    fonts = filter_if_exists(fonts, "stretch", stretch)
    fonts = filter_if_exists(fonts, "monospace", monospace)
    fonts = filter_if_exists(fonts, "opticalsize", opticalsize)
    fonts = filter_if_exists(fonts, "foundry", foundry)
    fonts = filter_if_exists(fonts, "special", special)
    # We want "italic" to fall back to "oblique".  We don't need the reverse.
    try:
        fonts = filter_if_exists(fonts, "style", style)
    except NoFontFoundError as e:
        if style == "italic":
            try:
                fonts = filter_if_exists(fonts, "style", "oblique")
            except NoFontFoundError as e2:
                raise e
        else:
            raise e
    # If there is more than one option, give the user a chance to narrow it down.
    if len(fonts) > 1 and not multiple:
        differing_props = []
        for prop in ["family_name", "weight", "weight_number", "stretch", "monospace",
                     "opticalsize", "style", "foundry", "special"]:
            allprops = set(f[prop] for f in fonts)
            if len(allprops) > 1:
                if prop == "weight_number":
                    prop = "weight"
                differing_props.append((prop, allprops))
        if len(differing_props) > 0:
            errortext = "Multiple matching fonts, please specify one or more of the following:\n"
            errortext += "\n".join([f"    '{prop}' is one of: {sorted(vals)}"
                                    for prop,vals in differing_props])
            raise MultipleFontsFoundError(errortext)
        else:
            fonts = sorted(fonts, key=lambda x : x['fname'])
            warningtext = "Warning: Multiple versions of the specified font are installed:\n"
            for f in fonts:
                warningtext += "    " + f['fname'] + "\n"
            warningtext += "We are arbitrarily choosing:"
            warningtext += "    " + fonts[0]['fname'] + "\n"
            matchingfonts = fonts[0:1]
            print(warningtext)
    if multiple:
        loadedttf = fonts
    else:
        loadedttf = fonts[0]
    _find_font_cache[cachename] = loadedttf
    return loadedttf

_find_font_family_cache = {}
def find_font_family(name, *, stretch=None, opticalsize=None, monospace=None, foundry=None, special=None):
    global _find_font_family_cache
    cachename = (name, stretch, opticalsize, monospace, foundry, special)
    if cachename in _find_font_family_cache.keys():
        return _find_font_family_cache[cachename]
    # Get all the fonts which match the user's search criteria
    allfonts = find_font(name, stretch=stretch, opticalsize=opticalsize, monospace=monospace, foundry=foundry, special=special, multiple=True)
    # Check to make sure that only one family was returned.
    familynames = list(sorted(set([f['family_name'] for f in allfonts])))
    if len(familynames) > 1:
        errortext = "Please be more specific in specifying font family.\nSpecify one of the following font names:\n    "
        errortext += ", ".join(f'"{f}"' for f in familynames)
        foundries = list(sorted(set([f['foundry'] for f in allfonts])))
        if len(foundries) > 1:
            errortext += "\nor one of the following foundries:\n    "
            errortext += ", ".join(f'"{f}"' for f in foundries)
        raise MultipleFontsFoundError(errortext)
    # Now we try to prune everything down to just one font.  We start by making
    # sure there is only one stretch and optical size.
    opticalsizes = set(f['opticalsize'] for f in allfonts)
    if len(opticalsizes) > 1:
        errortext = "Please specify an optical size using the function argument opticalsize=[value].  Valid values for this font are:\n    "
        errortext += ", ".join(f'"{os}"' for os in opticalsizes)
        raise MultipleFontsFoundError(errortext)
    stretches = set(f['stretch'] for f in allfonts)
    if len(stretches) > 1:
        errortext = "Please specify a stretch using the function argument stretch=[value].  Valid values for this font are:\n    "
        errortext += ", ".join(f'"{s}"' for s in stretches)
        raise MultipleFontsFoundError(errortext)
    def find_member(allfonts, bold, italic):
        basestyle = 0
        if bold:
            basestyle += 2
            weights = ["bold", "semibold"]
        else:
            weights = ["roman", "normal", "book", "regular", "medium"]
        if italic:
            basestyle += 1
            styles = ['italic', 'oblique']
        else:
            styles = ['normal']
        # Now let's try to find the base font for this family.  We will do this
        # by slowly pruning the list down to only one font.  The "base_style"
        # should be 0 for the base font, so we check that first.
        matchingfonts = [f for f in allfonts if f['base_style'] == basestyle]
        # If that didn't narrow it down, first filter out all of the italic fonts,
        # assuming there is at least one non-italic font.
        if len(matchingfonts) > 1 and len(set(f['style'] for f in matchingfonts)) > 1:
            # This would fail if a font family only had italic and oblique with a
            # base_style of 0, but the chance of a font designer screwing up so
            # badly is very low.
            matchingfonts = [f for f in matchingfonts if f['style'] in styles]
        # Next, check to see if there are multiple version of the font
        # installed.  If so, pick the most recent version.  This is a crude
        # method of choosing the most recent version, but it is the best easy
        # solution given the heterogeneity in naming conventions for the
        # version string across fonts.
        versions = list(sorted(set(f['version'] for f in matchingfonts)))
        if len(matchingfonts) > 1 and len(versions) > 1:
            matchingfonts = [f for f in matchingfonts if f['version'] == versions[-1]]
        # If things still aren't working, now we check the foundry for
        # duplicates.  We hold off on checking the foundry information for as
        # long as possible (instead of doing it above) because it is often
        # inconsistent.
        foundries = list(sorted(set(f['foundry'] for f in matchingfonts)))
        if len(matchingfonts) > 1 and len(foundries) > 1:
            errortext = "Please specify a foundry using the function argument foundry=[value].  Valid values for this font are:\n    "
            errortext += ", ".join(f'"{f}"' for f in foundries)
            raise MultipleFontsFoundError(errortext)
        # Now do the same thing for the special property, which can also be
        # inconsistent.
        specials = list(sorted(set(f['special'] for f in matchingfonts)))
        if len(matchingfonts) > 1 and len(specials) > 1:
            errortext = "Please specify a special property using the function argument special=[value].  Valid values for this font are:\n    "
            errortext += ", ".join(f'"{s}"' for s in specials)
            raise MultipleFontsFoundError(errortext)
        # If we still have too many fonts, start to filter based on the name of the
        # font's weight.
        if len(matchingfonts) > 1 and len(set(f['weight'] for f in matchingfonts)) > 1:
            # We loop through these because it allows us to prioritize the first
            # ones, e.g. if a font contains both "medium" and "book".
            for weight in weights:
                newmatchingfonts = [f for f in matchingfonts if f['weight'] == weight]
                if newmatchingfonts:
                    matchingfonts = newmatchingfonts
                    break
        # If there are multiple matching fonts, check to see if they have the
        # same identifier.  If so, they are the same font and we can choose
        # either arbitrarily.  For consistency, we sort them first based on
        # filename.
        if len(matchingfonts) > 1 and len(set(f['identifier'] for f in matchingfonts)) == 1:
            matchingfonts = sorted(matchingfonts, key=lambda x : x['fname'])
            matchingfonts = matchingfonts[0:1]
        # Now we really should only have one font.  If not, it's an error and we
        # should alert the user.
        if len(matchingfonts) > 1:
            matchingfonts = sorted(matchingfonts, key=lambda x : x['fname'])
            warningtext = "Warning: Multiple versions of the specified font are installed:\n"
            for f in matchingfonts:
                warningtext += "    " + f['fname'] + "\n"
            warningtext += "We are arbitrarily choosing:"
            warningtext += "    " + matchingfonts[0]['fname'] + "\n"
            matchingfonts = matchingfonts[0:1]
            print(warningtext)
        _find_font_family_cache[cachename] = matchingfonts[0]
        return matchingfonts[0]
    styles = {}
    try:
        styles['regular'] = find_member(allfonts, bold=False, italic=False)
    except:
        print("Warning, regular font not found")
    try:
        styles['italic'] = find_member(allfonts, bold=False, italic=True)
    except:
        print("Warning, italic font not found")
    try:
        styles['bold'] = find_member(allfonts, bold=True, italic=False)
    except:
        print("Warning, bold font not found")
    try:
        styles['bolditalic'] = find_member(allfonts, bold=True, italic=True)
    except:
        print("Warning, bold-italic font not found")
    for style in styles:
        if styles[style]['num_faces'] > 1:
            print(f"Warning, font is a .ttc file with {styles[style]['num_faces']} embedded faces.  CanD can only use one.  If you have problems, try using a ttf or otf font instead.")
            break
    _find_font_family_cache[cachename] = styles
    return styles

#TODO: Neutraface, EB Garamond, penumbra, gotham, fell, stone sans, univers arkandis (chooses condensed by default), Bell
# Universalis is has the wrong value for typographic subfamily
# Fell has crazy labeling



import paranoid as pns
import numpy as np
import matplotlib
import matplotlib.figure
from PIL import Image, PngImagePlugin
import fitz as mupdf # PyMuPDF
import tempfile
import atexit
import os
from .metrics import Metric, Vector, Point, BinopPoint, BinopVector, Height, Width
from .fontant import find_font, find_font_family, MultipleFontsFoundError, NoFontFoundError
from ._version import __version__

_idstr = f"CanD {__version__} (github.com/mwshinn/cand)"

# If IPython is installed, try to import the display code for it.
try:
    from IPython.display import Image as IPython_Image, display as IPython_display
except ImportError:
    pass

@pns.paranoidclass
class Canvas:
    """Canvas is a convenient way of arranging and organizing axes and other elements on a figure.

    Create a Canvas by specifying its x and y size in inches (`size_x`
    and `size_y`).  Optionally, specify the font size in points
    (`fontsize`) and the font name as a valid postscript font (`font`).

    A Canvas makes it easy to add create attractive composite layouts
    in matplotlib.  When a figure consists of just a single axis, it
    is not difficult to create an attractive layout.  However, when
    many axes are combined (e.g. into subpanels), it becomes more
    difficult to standardize the components such as font size and
    style.  Additionally, for non-grid layouts, it is difficult to
    correctly compute the position of the axes, align text and arrows,
    and other layout-related tasks.  Canvas provides a set of features
    to make layouts and fonts trivial.

    """
    @pns.accepts(pns.Self, pns.Number, pns.Number, pns.String)
    @pns.paranoidconfig(unit_test=False)
    def __init__(self, size_x, size_y, unit="inches"):
        self.axes = dict()
        self.default_unit = "figure"
        self.fontsize = 8
        self.fontsize_ticks = 8
        self.fontsize_title = 8
        self.images = []
        self.tmpfiles = []
        self.font = dict(name="DejaVu Sans", stretch="normal")
        atexit.register(self._cleanup)
        self.localRc = {}
        
        self.backend = "default"
        # Create default units.  Dictionary of tuples indexed by unit
        # name.  First two elements are x and y scale (with respect to
        # inches) and the last is the origin of the coordinate system.
        if unit in ["inches", "in", "inch"]:
            size_x_inches = size_x
            size_y_inches = size_y
        elif unit in ["cm", "centimeter", "centimeters", "centimetre", "centimetres"]:
            size_x_inches = size_x/2.54
            size_y_inches = size_y/2.54
        elif unit in ["mm", "millimeter", "millimeters", "millimetre", "millimetres"]:
            size_x_inches = size_x/25.4
            size_y_inches = size_y/25.4
        else:
            raise ValueError("Invalid unit")

        # Create matplotlib figure object
        figsize = (size_x_inches, size_y_inches)
        self.figure = matplotlib.figure.Figure(figsize=figsize)
        self.size = figsize # Size of the figure in inches
        self.trans_absolute = self.figure.dpi_scale_trans

        self.units = dict()
        self.add_unit("in", Vector(1/size_x_inches, 1/size_y_inches, "figure"))
        self.units["inch"] = self.units["in"]
        self.units["inches"] = self.units["in"]
        self.add_unit("cm", Vector(1/2.54, 1/2.54, "inches"))
        self.units["centimeters"] = self.units["cm"]
        self.units["centimeter"] = self.units["cm"]
        self.units["centimetres"] = self.units["cm"]
        self.units["centimetre"] = self.units["cm"]
        self.add_unit("mm", Vector(.1, .1, "cm"))
        self.units["millimeter"] = self.units["mm"]
        self.units["millimeters"] = self.units["mm"]
        self.units["millimetre"] = self.units["mm"]
        self.units["millimetres"] = self.units["mm"]
        self.add_unit("pt", Vector(1/72, 1/72, "inches"))
        self.units["point"] = self.units["pt"]
        self.units["points"] = self.units["pt"]
        self.add_unit("px", Vector(1/self.figure.dpi, 1/self.figure.dpi, "in"))
        self.units["pixel"] = self.units["px"]
        self.units["pixels"] = self.units["px"]
    def _cleanup(self):
        for f in self.tmpfiles:
            os.remove(f)
    def set_font(self, name, *, size=None, weight=None, style=None, stretch=None, foundry=None, special=None, opticalsize=None, monospace=None, ticksize=None, titlesize=None):
        if size:
            # Set up font sizes
            self.localRc['font.size'] = size
            self.localRc['axes.titlesize'] = titlesize or size
            self.localRc['axes.labelsize'] = size
            self.localRc['xtick.labelsize'] = ticksize or size
            self.localRc['ytick.labelsize'] = ticksize or size
            self.localRc['legend.fontsize'] = size
            self.localRc['figure.titlesize'] = titlesize or size
            self.fontsize = size
            self.fontsize_ticks = ticksize or size
            self.fontsize_title = titlesize or size
        newfont = {}
        newfont['name'] = name
        if weight is not None:
            newfont['weight'] = weight
        if style is not None:
            newfont['style'] = style
        if stretch is not None:
            newfont['stretch'] = stretch
        if foundry is not None:
            newfont['foundry'] = foundry
        if special is not None:
            newfont['special'] = special
        if monospace is not None:
            newfont['monospace'] = monospace
        if opticalsize is not None:
            newfont['opticalsize'] = opticalsize
        print(newfont)
        try:
            self._get_font(**newfont) # Run once to confirm the font is valid
        except MultipleFontsFoundError as e:
            e.args = ("Font specified in call to Canvas.set_font() was not specific enough.\n"+e.args[0],) + e.args[1:]
            raise
        self.font = newfont
        
    def _get_font(self, name=None, *, size=None, weight=None, style=None, stretch=None, foundry=None, special=None, opticalsize=None, monospace=None):
        if name is None:
            defaultfont = self.font.copy()
        else:
            defaultfont = {}
        # Otherwise, go through and make any necessary changes to the default
        # Canvas font.
        if name is not None:
            defaultfont['name'] = name
        if weight is not None:
            defaultfont['weight'] = weight
        if style is not None:
            defaultfont['style'] = style
        if stretch is not None:
            defaultfont['stretch'] = stretch
        if foundry is not None:
            defaultfont['foundry'] = foundry
        if special is not None:
            defaultfont['special'] = special
        if monospace is not None:
            defaultfont['monospace'] = monospace
        if opticalsize is not None:
            defaultfont['opticalsize'] = opticalsize
        if size is None:
            size = self.fontsize
        # If we just want the bold or italic versions of a font, we can use the
        # font family method.
        if ('weight' not in defaultfont.keys() or defaultfont['weight'] == "bold") and \
           ('style' not in defaultfont.keys() or defaultfont['style'] in ["italic", "oblique"]):
            family = find_font_family(name=defaultfont['name'], stretch=defaultfont.get('stretch', None),
                                     foundry=defaultfont.get('foundry', None), special=defaultfont.get('special', None), opticalsize=defaultfont.get('opticalsize', None),
                                      monospace=defaultfont.get('monospace', None))
            if 'weight' not in defaultfont.keys() and 'style' not in defaultfont.keys():
                key = "regular"
            elif 'weight' in defaultfont.keys() and 'style' not in defaultfont.keys():
                key = "bold"
            elif 'weight' not in defaultfont.keys() and 'style' in defaultfont.keys():
                key = "italic"
            else:
                key = "bolditalic"
            if key in family.keys():
                fprops = matplotlib.font_manager.FontProperties(fname=family[key]['fname'], size=size)
                return fprops
        # If not, we use the more heavy duty font selection tool
        fontfile = find_font(**defaultfont)['fname']
        fprops = matplotlib.font_manager.FontProperties(fname=fontfile, size=size)
        return fprops
    @pns.accepts(pns.Self, pns.Maybe(pns.String), pns.String)
    def use_latex(self, preamble=None, engine="pdflatex"):
        """Use latex to render all text.

        `engine` is the latex engine ("texsystem" in matplotlib lingo)
        used to render the text.  It can be "pdflatex", "lualatex", or
        "xelatex".

        Note that this will break math font compatibility.  Because
        latex is being used to render the fonts, you must import a
        package that allows latex to render in the desired font.  For
        example, if `engine` is "pdflatex", to get Helvetica, set
        preamble=r"\\usepackage[scaled]{helvet}\\usepackage[helvet]{sfmath}".
        """
        self.backend = "latex"
        self.latex_engine = engine
        self.latex_preamble = preamble
    @pns.accepts(pns.Self, pns.String, Vector, Point)
    @pns.ensures('not self.is_valid_identifier(name)')
    def add_unit(self, name, scale, origin=Point(0, 0, "absolute")):
        """Create a new unit of measure.

        A unit is defined by an affine transformation.  It has an
        origin and a scale in the x and y dimensions.  The name of the
        unit is given by `name`, the scale in the x and y directions
        is given by `scale`, and the origin of the coordinate system
        is optionally specified by `origin`.

        """
        assert self.is_valid_identifier(name), f"Invalid unit name {name!r}"
        scale = self.convert_to_absolute_length(scale)
        origin = self.convert_to_absolute_coord(origin)
        self.units[name] = (scale.width().x, scale.height().y, origin)
    @pns.accepts(pns.Self, pns.String)
    def set_default_unit(self, name):
        """Changes the default unit for the Canvas.

        When a unit is not specified for a Point or a Vector, it uses
        the default unit of the Canvas.  The default is initially set
        to figure coordinates.

        """
        assert self.is_unit(name), f"Invalid unit name {name!r} set as default"
        self.default_unit = name
    @pns.accepts(pns.Self, pns.String, Point, Point)
    @pns.ensures("not self.is_valid_identifier(name)")
    def add_axis(self, name, pos_ll, pos_ur):
        """Create a new axis on the Canvas.

        Create a new matplotlib axis named `name`, with the lower left
        corner at point `pos_ll` and the upper right corner at
        `pos_ur`.  Note that the axis labels, title, and other
        elements may fall outside of this bounding box.  The axis of a
        Canvas object "c" with name "axname" can be accessed by:

            ax = c.ax("axname")

        Then, ax can be used like a normal matplotlib axis.

        This also creates two new coordinate systems: one is given by
        the axis' name, which uses the coordinates of the data in the
        axis, and "axis_" prepended to the axis' name, which is (0,0)
        at the lower left corner and (1,1) at the upper right corner

        """
        assert name not in self.axes.keys(), "Axis name alredy exists"
        assert self.is_valid_identifier(name), f"Invalid axis name {name!r}"
        assert name != "figure", "Invalid axis name 'figure'"
        # Need to use figure coordinates here because of a bug in
        # matplotlib which won't accept the 'transform' argument to
        # add_axes.
        pt_ll = self.convert_to_figure_coord(pos_ll)
        pt_ur = self.convert_to_figure_coord(pos_ur)
        ax = self.figure.add_axes([pt_ll.x, pt_ll.y, pt_ur.x-pt_ll.x, pt_ur.y-pt_ll.y], label=name)
        self.axes[name] = ax
        return ax
    @pns.accepts(pns.Self, pns.String)
    @pns.ensures("self.is_unit(name)")
    def ax(self, name):
        """Return the axis of name `name`."""
        return self.axes[name]
    #@pns.accepts(pns.Self, pns.String, Point, Point, pns.Or(pns.Tuple(pns.Number, pns.Number), matplotlib.colors.Normalize))
    def add_colorbar(self, name, pos_ll, pos_ur, bounds, **kwargs):
        """Add a colorbar.

        Name the colorbar `name`, which will be an axis just like any
        other Canvas axis.  The lower left corner is at `pos_ll` and
        the upper right corner is at `pos_ur`.  `bounds` should be a
        tuple of numbers, representing the lower and upper bounds of
        the colorbar, respectively.  All other arguments are passed
        directly to matplotlib.colorbar.ColorbarBase.  Of particular
        note is the argument `cmap`, defining the colormap.

        Return the ColorbarBase object.
        """

        ax = self.add_axis(name, pos_ll, pos_ur)
        if isinstance(bounds, tuple):
            norm = matplotlib.colors.Normalize(vmin=bounds[0], vmax=bounds[1])
        else:
            norm = bounds
        size = self.convert_to_absolute_coord(pos_ur - pos_ll)
        orientation = "horizontal" if size.x > size.y else "vertical"
        if 'cmap' in kwargs:
            kwargs['cmap'] = matplotlib.cm.get_cmap(kwargs['cmap'])
        colorbar = matplotlib.colorbar.ColorbarBase(ax, norm=norm, orientation=orientation, **kwargs)
        return colorbar
    @pns.accepts(pns.Self, pns.String)
    @pns.returns(pns.Boolean)
    def is_unit(self, name):
        """Whether the value is a in use as a unit of measure.
        
        `name` is a name to test.  Returns True if it is in use or
        False if it is not.

        """
        if name in ["default", "figure", "absolute", "-absolute", "various"]:
            return True
        if name in self.axes.keys():
            return True
        if name.startswith("axis_") and self.is_unit(name[5:]):
            return True
        if name in self.units.keys():
            return True
        return False
    @pns.accepts(pns.Self, pns.String)
    @pns.returns(pns.Boolean)
    def is_valid_identifier(self, name):
        """Whether the value can be used as an axis or unit name.
        
        `name` is a name to test.  Returns True if it can be used or
        False if it cannot (e.g. it is a reserved keyword).

        """
        if name.startswith("axis_"):
            return False
        return not self.is_unit(name)
    @pns.accepts(pns.Self, Metric)
    @pns.returns(Metric)
    @pns.ensures("return.coordinate == 'absolute'")
    def convert_to_absolute_coord(self, point):
        """Convert the coordinate system of the Point or Vector `point` to be "figure".

        We can convert any coordinate system to the "figure"
        coordinate system for a given Point or Vector.  This is useful
        for comparing Points and Vectors, and also used internally as
        a universal coordinate system.  It also collapses Binop
        objects into Points or Vectors.

        """
        if isinstance(point, Vector):
            return self.convert_to_absolute_length(point)
        if point.coordinate == "default":
            return self.convert_to_absolute_coord(Point(point.x, point.y, self.default_unit))
        if point.coordinate == "absolute":
            return point
        if point.coordinate == "figure":
            return Point(point.x*self.size[0], point.y*self.size[1], "absolute")
        if point.coordinate == "-absolute":
            return Point(self.size[0]-point.x, self.size[1]-point.y, "absolute")
        if point.coordinate in ["Msize", "fontsize"]: # Msize for backward compatibility
            return self.convert_to_absolute_coord(Point(point.x*self.fontsize, point.y*self.fontsize, "point"))
        if point.coordinate in self.axes.keys():
            # The call to autoscale_view fix the problem that automatic data
            # limits are updated lazily, and thus, gives an outdated transData
            # matrix until a display function is called.
            self.axes[point.coordinate].autoscale_view()
            tf_data = self.axes[point.coordinate].transData
            tf_fig = self.trans_absolute.inverted()
            x,y = tf_fig.transform(tf_data.transform((point.x, point.y)))
            return Point(x, y, "absolute")
        if point.coordinate.startswith("axis_") and point.coordinate[5:] in self.axes.keys():
            tf_ax = self.axes[point.coordinate[5:]].transAxes
            tf_fig = self.trans_absolute.inverted()
            x,y = tf_fig.transform(tf_ax.transform((point.x, point.y)))
            return Point(x, y, "absolute")
        if point.coordinate in self.units:
            u = point.coordinate
            return Vector(point.x*self.units[u][0], point.y*self.units[u][1], "absolute") + self.units[u][2]
        if isinstance(point, BinopPoint):
            # Call recursively, but handle the scalar case separately
            if isinstance(point.lhs, Point) or isinstance(point.lhs, Vector):
                lhs = self.convert_to_absolute_coord(point.lhs)
            else:
                lhs = point.lhs
            if isinstance(point.rhs, Point) or isinstance(point.rhs, Vector):
                rhs = self.convert_to_absolute_coord(point.rhs)
            else:
                rhs = point.rhs
            return BinopPoint.op_table[point.op](lhs, rhs)
        raise ValueError("Invalid point coordinate system %s" % point.coordinate)
    @pns.accepts(pns.Self, Vector)
    @pns.returns(Vector)
    @pns.ensures("return.coordinate == 'absolute'")
    def convert_to_absolute_length(self, vector):
        """Convert the coordinate system of the Vector `vector` to be "figure".

        We can convert any coordinate system to the "figure"
        coordinate system for a given Vector.  This is useful for
        comparing vectors, and also used internally as a universal
        coordinate system.  It also collapses BinopVectors into
        Vectors.

        """
        if vector.coordinate == "absolute":
            return vector
        elif isinstance(vector, BinopVector):
            # Call recursively, but handle the scalar case separately
            if isinstance(vector.lhs, Point) or isinstance(vector.lhs, Vector):
                lhs = self.convert_to_absolute_coord(vector.lhs)
            else:
                lhs = vector.lhs
            if isinstance(vector.rhs, Point) or isinstance(vector.rhs, Vector):
                rhs = self.convert_to_absolute_coord(vector.rhs)
            else:
                rhs = vector.rhs
            return BinopVector.op_table[vector.op](lhs, rhs)
        else:
            origin = Point(0, 0, vector.coordinate)
            return self.convert_to_absolute_coord(origin+vector) - self.convert_to_absolute_coord(origin)
    @pns.accepts(pns.Self, Metric)
    @pns.returns(Metric)
    @pns.ensures("return.coordinate == 'figure'")
    def convert_to_figure_coord(self, point):
        """Converts units in a Point or Vector to "figure"."""
        if isinstance(point, Vector):
            return self.convert_to_figure_length(point)
        p = self.convert_to_absolute_coord(point)
        return Point(p.x / self.size[0], p.y / self.size[1], "figure")
    @pns.accepts(pns.Self, Vector)
    @pns.returns(Vector)
    @pns.ensures("return.coordinate == 'figure'")
    def convert_to_figure_length(self, vec):
        """Converts units in a Vector to "figure"."""
        v = self.convert_to_absolute_length(vec)
        return Vector(v.x / self.size[0], v.y / self.size[1], "figure")
    @pns.accepts(pns.Self, pns.List(Point))
    def add_poly(self, points, **kwargs):
        """Draw a polygon with given vertices.

        Vertices are passed as a list of Point objects via the
        `points` argument.  All other keyword arguments are passed
        directly to matplotlib.patches.Polygon.

        """
        np_points = np.zeros((len(points), 2))
        for i,p in enumerate(points):
            pt = self.convert_to_absolute_coord(p)
            np_points[i] = [pt.x, pt.y]
        if "fill" not in kwargs.keys():
            kwargs['fill'] = False
        poly = matplotlib.patches.Polygon(np_points, transform=self.trans_absolute, **kwargs)
        self.figure.add_artist(poly)
    @pns.accepts(pns.Self, Point, Point)
    def add_rect(self, pos_ll, pos_ur, **kwargs):
        """Draw a rectangle.

        The lower left corner is the Point `pos_ll` and the upper
        right corner is the Point `pos_ur`.  All other keyword
        arguments are passed directly to matplotlib.patches.Polygon.

        """
        pt_ll = self.convert_to_absolute_coord(pos_ll)
        pt_ur = self.convert_to_absolute_coord(pos_ur)
        connect = pt_ur - pt_ll
        # When drawing a box you have to duplicate the last point for
        # some reason... probably a bug in matplotlib
        self.add_poly([pt_ll, pt_ll+connect.height(), pt_ur, pt_ll+connect.width(), pt_ll, pt_ll], **kwargs)
    @pns.accepts(pns.Self, Point, Point)
    def add_box(self, pos_ll, pos_ur, **kwargs):
        """Draw a bounding box.

        The lower left corner is the Point `pos_ll` and the upper
        right corner is the Point `pos_ur`.  All other keyword
        arguments are passed directly to
        matplotlib.patches.FancyBboxPatch.

        Note that this differs from add_rect because it uses the
        FancyBboxPatch rather than matplotlib.patches.Polygon.  Thus,
        this function can be used for, e.g., boxes with round corners
        """
        pt_ll = self.convert_to_absolute_coord(pos_ll)
        pt_ur = self.convert_to_absolute_coord(pos_ur)
        diff = pt_ur - pt_ll
        if "fill" not in kwargs.keys():
            kwargs['fill'] = False
        if 'mutation_scale' not in kwargs.keys():
            kwargs['mutation_scale'] = 1/2 # 1/2 inch mutation scale
        box = matplotlib.patches.FancyBboxPatch((pt_ll.x, pt_ll.y), diff.x, diff.y, transform=self.trans_absolute, **kwargs)
        self.figure.add_artist(box)
    @pns.accepts(pns.Self, Point, Point)
    def add_ellipse(self, pos_ll, pos_ur, **kwargs):
        """Draw an ellipse.

        The lower left corner is the Point `pos_ll` and the upper
        right corner is the Point `pos_ur`.  Note that rotation is not
        currently possible because it is not clear in which coordinate
        system the rotation would be applied.  All other keyword
        arguments are passed directly to matplotlib.patches.Ellipse.
        """
        pt_ll = self.convert_to_absolute_coord(pos_ll)
        pt_ur = self.convert_to_absolute_coord(pos_ur)
        diff = pt_ur - pt_ll
        center = pt_ll | pt_ur
        if "angle" in kwargs.keys():
            print("Warning: the 'angle' keyword passed to add_ellipse may give unexpected results.")
        # When drawing a box you have to duplicate the last point for
        # some reason... probably a bug in matplotlib
        e = matplotlib.patches.Ellipse(xy=tuple(center), width=diff.width().x, height=diff.height().y,
                                       transform=self.trans_absolute, **kwargs)
        self.figure.add_artist(e)
    def add_arrow(self, frm, to, arrowstyle="->,head_width=3,head_length=4", lw=2, linestyle='solid', **kwargs):
        """Draw an arrow.

        Draw an arrow from Point `frm` to Point `to`.  All other
        keyword arguments are passed directly to
        matplotlib.patches.FancyArrowPatch.  Reasonable default
        arguments are given, but these can be overridden.

        """
        pt_frm = self.convert_to_absolute_coord(frm)
        pt_to = self.convert_to_absolute_coord(to)
        pt_delta = pt_frm - pt_to
        if "linewidth" in kwargs:
            lw = kwargs['linewidth']
        # if "shrinkA" not in kwargs:
        #     kwargs['shrinkA'] = 0
        # if "shrinkB" not in kwargs:
        #     kwargs['shrinkB'] = 0
        arrow = matplotlib.patches.FancyArrowPatch(tuple(pt_frm), tuple(pt_to), transform=self.trans_absolute,
                                                       arrowstyle=arrowstyle, lw=lw, linestyle=linestyle, **kwargs)
        self.figure.patches.append(arrow)
    def add_text(self, text, pos, font=None, size=None, weight=None, style=None, stretch=None, foundry=None, special=None, opticalsize=None, monospace=None, **kwargs):
        """Add text at a given point.

        Draw the text `text` at Point `pos`.  The argument `weight`
        specifies the font weight, which varies depending on the
        weights of the font, but potential options may be "light",
        "roma", "bold", "heavy", "black".  The size is given by `size`
        in points, which defaults to the Canvas' default. The width is
        given by "stretch", which can be "consensed", "normal", or
        "wide", depending on the font.  All other keyword arguments
        are passed to matplotlib.pyplot.text.  Notably, the
        `horizontalalignment` and `verticalalignment` arguments are
        often useful.  (`ha` and `va` are also accepted as aliases for
        matplotlib compatibility).
        """
        if size is None:
            size = self.fontsize
        kwargs = kwargs.copy()
        if 'horizontalalignment' not in kwargs:
            kwargs['horizontalalignment'] = kwargs['ha'] if 'ha' in kwargs else 'center'
        if 'verticalalignment' not in kwargs:
            kwargs['verticalalignment'] = kwargs['va'] if 'va' in kwargs else 'center'
        if 'fontname' in kwargs.keys() and font is None:
            font = kwargs['fontname']
        pt = self.convert_to_absolute_coord(pos)
        # Check valid font names with matplotlib.font_manager.FontManager().findfont(name)
        #self.figure.text(pt.x, pt.y, text, transform=self.figure.transFigure, fontdict={'fontname': "Helvetica Neue LT Std", "fontsize": 20})
        fprops = self._get_font(name=font, weight=weight, size=size,
                                stretch=stretch, style=style, foundry=foundry, special=special,
                                opticalsize=opticalsize, monospace=monospace)
        self.figure.text(pt.x, pt.y, text, transform=self.trans_absolute,
                         fontproperties=fprops, fontsize=size, **kwargs)
    def add_line(self, frm, to, **kwargs):
        """Draw a line.

        Draw a line from Point `frm` to Point `to`.  All other keyword
        arguments are passed directly to matplotlib.lines.Line2D.

        """
        frm = self.convert_to_absolute_coord(frm)
        to = self.convert_to_absolute_coord(to)
        l2d = matplotlib.lines.Line2D([frm.x, to.x], [frm.y, to.y],
                                      transform=self.trans_absolute, **kwargs)
        self.figure.add_artist(l2d)
    def add_marker(self, pos, **kwargs):
        """Draw a matplotlib marker.

        Plot a marker at Point `pos`.  All other keyword arguments are
        passed directly to matplotlib.lines.Line2D.

        """
        pos = self.convert_to_absolute_coord(pos)
        l2d = matplotlib.lines.Line2D([pos.x], [pos.y], transform=self.trans_absolute, **kwargs)
        self.figure.add_artist(l2d)
    @pns.accepts(pns.Self, Point, pns.List(pns.Tuple(pns.String, pns.Dict(k=pns.String, v=pns.Unchecked))), pns.Maybe(pns.Natural1), Metric, Metric, Metric)
    def add_legend(self, pos_tl, els, fontsize=None, line_spacing=Height(2.2, "Msize"), sym_width=Width(2.3, "Msize"), padding_sep=Width(1.2, "Msize")):
        """Add a legend without using the matplotlib API.

        The top-left corner of the legend should be located at the
        Point `pos_tl`.  The `els` argument should be a list of tuples
        representing the elements to include in the legend.  The first
        element of each tuple should be the name of the legend item,
        the second element should be a dictionary of line properties
        to be passed to the add_line and add_marker functions.  To
        withhold drawing a line and only draw markers, set 'linestyle'
        to the string 'None'.

        Additional parameters control formatting.  `line_spacing`
        determines spacing between each line of descriptive text in
        the legend.  `sym_width` is the width of the symbols (lines
        and markers). `padding_sep` is the separation between the
        symbols and the descriptive text.
        """
        if fontsize is None:
            fontsize = self.fontsize
        pos_tl = self.convert_to_absolute_coord(pos_tl)
        assert len(els) >= 1
        # Get the text height
        fprops = self._get_font()
        # All params are in units of M width or height
        padding_top = Height(0, "Msize") # Space on top of figure
        padding_left = Width(0, "Msize") # Space on left of lines
        # Convert these to an easier coordinate system
        padding_top = self.convert_to_absolute_length(padding_top)
        padding_left = self.convert_to_absolute_length(padding_left)
        padding_sep = self.convert_to_absolute_length(padding_sep)
        line_spacing = self.convert_to_absolute_length(line_spacing)
        sym_width = self.convert_to_absolute_length(sym_width)
        top_left = pos_tl - padding_top + padding_left
        for i in range(0, len(els)):
            # Figure out the vertical position of this element of the legend
            y_offset = -1*line_spacing*i
            # Draw the text
            self.add_text(els[i][0],
                          top_left + sym_width + padding_sep + y_offset,
                          horizontalalignment="left", size=fontsize)
            pt1 = top_left + y_offset
            pt2 = top_left + sym_width + y_offset
            # Draw the line
            params_nomarker = els[i][1].copy()
            params_nomarker['markersize'] = 0
            self.add_line(pt1, pt2, **params_nomarker)
            # Draw the marker.  We need this in an if statement due to
            # a bug in matplotlib.
            params_noline = els[i][1].copy()
            params_noline['linestyle'] = 'None'
            self.add_marker(pt1+(pt2-pt1)/2, **params_noline)
    @pns.accepts(pns.Self, pns.List(pns.Or(pns.Tuple(pns.String, pns.String),
                                           pns.Tuple(pns.String, pns.String, Metric))),
                 pns.Natural0)
    def add_figure_labels(self, labs, size=12):
        """Add letter labels to axes.

        `labs` should be a list of tuples of length 2 or 3.  The first
        element of each tuple is a string of the label text, most
        commonly a single lowercase letter.  The second element of the
        tuple is the name of an axis on the canvas.  The optional
        final element may be either a point (to specify manual
        positioning) or a vector (to specify an offset from the
        default position).
        """

        for l in labs:
            offset = Vector(-.15, .12, "absolute")
            unit_name = "axis_"+l[1] if l[1] in self.axes.keys() else l[1]
            assert self.is_unit(unit_name), f"Invalid unit or axis {l[1]} in figure label"
            loc = Point(0, 1, unit_name) + offset
            if len(l) == 3:
                if isinstance(l[2], Vector):
                    loc += l[2]
                elif isinstance(l[2], Point):
                    loc = l[2]
            self.add_text(l[0], loc, weight="bold", size=size)
    def fix_fonts(self):
        """Convert all text to the desired font.

        This function will fix all font objects within the figure.
        This must be called before displaying or saving the Canvas.
        Usually this is called automatically, but can be called
        manually as well.
        """
        fprops = self._get_font()
        try:
            fprops_bold = self._get_font(weight="bold")
        except (MultipleFontsFoundError, NoFontFoundError):
            fprops_bold = fprops
        try:
            fprops_it = self._get_font(style="italic")
        except (MultipleFontsFoundError, NoFontFoundError):
            fprops_it = fprops
        fprops_ticks = self._get_font(size=self.fontsize_ticks)
        fprops_title = self._get_font(size=self.fontsize_title)
        for ax in self.figure.axes:
            for label in ax.get_xticklabels():
                label.set_fontproperties(fprops_ticks)
            for label in ax.get_yticklabels():
                label.set_fontproperties(fprops_ticks)
            ax.set_xlabel(ax.get_xlabel(), fontproperties=fprops)
            ax.set_ylabel(ax.get_ylabel(), fontproperties=fprops)
            ax.set_title(ax.get_title(), fontproperties=fprops_title)
            if ax.legend_:
                for t in ax.legend_.texts:
                    t.set_fontproperties(fprops)
        # Get Helvetica for math as well
        self.localRc['mathtext.fontset'] = 'custom'
        self.localRc['mathtext.rm'] = fprops.get_fontconfig_pattern()
        self.localRc['mathtext.default'] = "rm"
        self.localRc['mathtext.bf'] = fprops_bold.get_fontconfig_pattern()
        self.localRc['mathtext.it'] = fprops_it.get_fontconfig_pattern()
        self.localRc['mathtext.cal'] = fprops.get_fontconfig_pattern()
        self.localRc['mathtext.tt'] = fprops.get_fontconfig_pattern()
        self.localRc['mathtext.sf'] = fprops.get_fontconfig_pattern()
        # Workaround for a bug in matplotlib where there are no
        # negative signs in front of axis labels in latex mode
        if self.backend == "latex":
            self.localRc['axes.unicode_minus'] = False
    def _grid_space(self, frm, to, spacing, count):
        dist = to-frm
        figsize = (dist-(count-1)*spacing)/count
        pos = []
        for i in range(0, count):
            base = frm+i*(figsize+spacing)
            pos.append((base, base+figsize))
        return pos
    @pns.accepts(pns.Self, pns.List(pns.Maybe(pns.String)), pns.Natural1, Point, Point, pns.Maybe(Vector), pns.Maybe(Vector), pns.Maybe(Vector), pns.Maybe(Vector), pns.Maybe(Vector), pns.Maybe(Vector), pns.Maybe(pns.String))
    @pns.requires("int(spacing_x is not None) + int(size_x is not None) + int(spacing is not None) + int(size is not None) == 1") # Exactly one of spacing_x, size_x, spacing, or size must be specified
    @pns.requires("int(spacing_y is not None) + int(size_y is not None) + int(spacing is not None) + int(size is not None) == 1") # Exactly one of spacing_y, size_y, spacing, or size must be specified
    def add_grid(self, names, nrows, pos_ll, pos_ur, spacing_x=None, spacing_y=None, spacing=None, size_x=None, size_y=None, size=None, unitname=None):
        """Create a grid of axes.

        Axes are specified by the `names` argument, a
        (one-dimensional) list of strings specifying the names of the
        axes to be created on the grid.  Optionally an element may be
        None, indicating that no axis should be created at this
        location on the grid.  Names start on the top line and go left
        to right and then line by line, just like words on a page.

        The argument `nrows` specifies how many rows should be
        created; the number of columns is implied by the number of
        rows and the length of `names`.

        The lower left corner of the grid is located as Point `pos_ll`
        and the upper right corner is at Point `pos_ur`.

        There are two ways to specify spacing of axes.  One method is
        to specify the size of the axes by the Vector `size`, or
        component-wise by the Width `size_x` and Height `size_y`.
        Axes will be automatically spaced to maintain the given size
        The other method is to specify the spacing of the axes with
        the Vector `spacing` or component-wise with the Width
        `spacing_x` and Height `spacing_y`.  The axes will be
        automatically sized to give the desired spacing.  Finally, it
        is possible to mix these styles by specifying `spacing_x` and
        `size_y` or vice versa.

        Optionally, `unitname` may specify a name for relative units
        based on the grid as a whole, i.e. where the origin is located
        at the bottom left corner of the bottom left figure in the
        grid, and (1,1) is located at the upper right corner of the
        upper right grid.
        """
        if spacing is not None:
            spacing_x = spacing.width()
            spacing_y = spacing.height()
        if size is not None:
            size_x = size.width()
            size_y = size.height()
        ncols = len(names)//nrows + int(len(names) % nrows != 0)
        pt_ll = self.convert_to_absolute_coord(pos_ll)
        pt_ur = self.convert_to_absolute_coord(pos_ur)
        if size_x is not None:
            size_x = self.convert_to_absolute_length(size_x)
            if ncols > 1:
                spacing_x = ((pt_ur - pt_ll).width() - size_x * ncols)/(ncols-1)
            elif ncols == 1:
                spacing_x = size_x*0
                w = (pt_ur-pt_ll).width()
                pt_ll = pt_ll + (w - size_x)/2
                pt_ur = pt_ur - (w - size_x)/2
        if size_y is not None:
            size_y = self.convert_to_absolute_length(size_y)
            if nrows > 1:
                spacing_y = ((pt_ur - pt_ll).height() - size_y * nrows)/(nrows-1)
            elif nrows == 1:
                spacing_y = size_y*0
                h = (pt_ur-pt_ll).height()
                pt_ll = pt_ll + (h - size_y)/2
                pt_ur = pt_ur - (h - size_y)/2
        spacing_x = self.convert_to_absolute_length(spacing_x)
        spacing_y = self.convert_to_absolute_length(spacing_y)
        posx = self._grid_space(pt_ll.x, pt_ur.x, spacing_x.x, ncols)
        posy = list(reversed(self._grid_space(pt_ll.y, pt_ur.y, spacing_y.y, nrows)))
        for i in range(0, len(names)):
            x = i % ncols
            y = i // ncols
            if names[i] is not None:
                self.add_axis(names[i], Point(posx[x][0], posy[y][0], "absolute"), Point(posx[x][1], posy[y][1], "absolute"))
        if unitname is not None:
            assert self.is_valid_identifier(unitname), f"Invalid axis name {unitname!r}"
            self.add_unit(unitname, (pt_ur-pt_ll), pt_ll)

    @pns.accepts(pns.Self, pns.String, pns.Maybe(pns.Natural1))
    def save(self, filename, dpi=600, *args, **kwargs):
        """Save the Canvas to a png or pdf file.

        The filename is specified by the string `filename`.  Optionally, the
        `dpi` argument may specify the dots per inch (dpi) of the output .png
        file, where larger numbers indicate a higher resolution and larger file
        size.  This is most relevant to .png output, but is also used when
        individual axes are rasterized through ax.set_rasterized().  Any
        additional arguments or keyword arguments are passed to the "savefig"
        function in matplotlib.

        """
        filetypes = ['png', 'pdf']
        filetype = next(ft for ft in filetypes if filename.endswith("."+ft))
        self.fix_fonts()
        # Force a white background in jupyter, which makes it transparent
        if self._in_jupyter():
            self.figure.patch.set_alpha(1)
        matplotlib.rc('pdf', fonttype=42) # This embeds (rather than subsets) all fonts in PDFs.
        # Lazy importing of matplotlib backend.  See https://matplotlib.org/3.3.1/tutorials/introductory/usage.html#the-builtin-backends
        if self.backend == "latex":
            from matplotlib.backends.backend_pgf import FigureCanvasPgf
            mplcanvas = FigureCanvasPgf(self.figure)
            matplotlib.rc('pgf', texsystem=self.latex_engine)
            if self.latex_preamble is not None:
                self.localRc['pgf.preamble'] = matplotlib.rcParams['pgf.preamble'] + self.latex_preamble
        elif self.backend == "default":
            from matplotlib.backends.backend_agg import FigureCanvasAgg
            mplcanvas = FigureCanvasAgg(self.figure)
        with matplotlib.rc_context(rc=self.localRc):
            self.figure.savefig(filename, dpi=dpi, *args, **kwargs)
        if filetype == "png":
            with Image.open(filename) as img:
                imgtext = img.text
                img = img.convert('RGBA')
                for image in self.images:
                    if image[0].endswith(".pdf"): # Convert pdf to png first
                        pdf = mupdf.open(image[0])
                        page = pdf[0]
                        imgpath = tempfile.mkstemp('.png')[1]
                        zoom = int(np.ceil(dpi/72)) if dpi else 1
                        page.getPixmap(alpha=True, matrix=mupdf.Matrix(zoom, zoom)).writeImage(imgpath)
                    else:
                        imgpath = image[0]
                    with Image.open(imgpath) as subimg:
                        subimg = subimg.convert('RGBA')
                        imwidth = img.size[0]
                        imheight = img.size[1]
                        pos_ll = image[1]
                        pos_ur = image[2]
                        # subimg.thumbnail(size)
                        bounds = (int(imwidth*pos_ll.x), int(imheight*(1-pos_ur.y)), int(imwidth*pos_ur.x), int(imheight*(1-pos_ll.y)))
                        subimg_size = (bounds[2]-bounds[0], bounds[3]-bounds[1])
                        subimg = subimg.resize(subimg_size, Image.LANCZOS)
                        img.alpha_composite(subimg, bounds[0:2])
                existing_meta = ("; "+imgtext['Software']) if 'Software' in imgtext.keys() else ""
                imgtext["Software"] = f"{_idstr}{existing_meta}"
                newmeta = PngImagePlugin.PngInfo()
                for k,v in imgtext.items():
                    newmeta.add_text(k, v)
                img.save(filename, pnginfo=newmeta)
        elif filetype == "pdf":
            pdf = mupdf.open(filename)
            page = pdf[0]
            pwidth = page.bound().width
            pheight = page.bound().height
            for image in self.images:
                pos_ll = image[1]
                pos_ur = image[2]
                rect = mupdf.Rect(pwidth*pos_ll.x, pheight*(1-pos_ur.y), pwidth*pos_ur.x, pheight*(1-pos_ll.y))
                #rect = mupdf.Rect(0.0, 0.0, 200.0, 200.0)
                #imagedoc = mupdf.open(image[0])
                #imagepdf_bytes = imagedoc.convertToPDF()
                #imagepdf = mupdf.open("pdf", imagepdf_bytes)
                #page.showPDFpage(rect, imagepdf)
                if image[0].endswith(".pdf"):
                    toinsert = mupdf.open(image[0])
                    page.showPDFpage(rect, src=toinsert, keep_proportion=False)
                else:
                    page.insertImage(rect, filename=image[0], keep_proportion=False)
            pdf.metadata['creator'] = f"{_idstr}; {pdf.metadata['creator']}"
            pdf.metadata['producer'] = f"{_idstr}; {pdf.metadata['producer']}"
            pdf.setMetadata(pdf.metadata)
            pdf.save(pdf.name, deflate=True, incremental=True, encryption=mupdf.PDF_ENCRYPT_KEEP)
            pdf.close()
    @pns.accepts(pns.Self, Vector, Point)
    def debug_grid(self, spacing, origin=Point(0, 0, "absolute"), **kwargs):
        """Create a grid to help you design a layout.

        This function is intended to help in the process of designing a layout,
        rather than as a feature of a finished figure.  It draws a grid over
        the top of the canvas at spacing defined by the Vector `spacing`.
        Optionally, it begins this grid at the Point `origin`.

        """
        bottom_left = Point(0, 0, "figure")
        top_right = Point(1, 1, "figure")
        args = {"zorder": 100, "alpha": .2, "c": "k"}
        args.update(kwargs)
        i = 0
        while True:
            xpos = i*spacing.width() + origin
            if self.convert_to_figure_coord(xpos).x >= 0:
                self.add_line(xpos >> bottom_left, xpos >> top_right, **args)
            else:
                break
            i -= 1
        i = 0
        while True:
            xpos = i*spacing.width() + origin
            if self.convert_to_figure_coord(xpos).x <= 1:
                self.add_line(xpos >> bottom_left, xpos >> top_right, **args)
            else:
                break
            i += 1
        i = 0
        while True:
            ypos = i*spacing.height() + origin
            if self.convert_to_figure_coord(ypos).y >= 0:
                self.add_line(bottom_left >> ypos, top_right >> ypos, **args)
            else:
                break
            i -= 1
        i = 0
        while True:
            ypos = i*spacing.height() + origin
            if self.convert_to_figure_coord(ypos).y <= 1:
                self.add_line(bottom_left >> ypos, top_right >> ypos, **args)
            else:
                break
            i += 1

    def add_image(self, filename, pos, unitname=None, height=None, width=None, horizontalalignment=None, verticalalignment=None, ha=None, va=None):
        """Add a png or pdf image to the Canvas.

        Insert a .png or .pdf file overlaid on the Canvas.  The string
        `filename` is the filename of the image, the Point `pos` is
        the location of the image, aligned according to
        `horizontalalignment` (may be "left", "center", or "right")
        and `verticalalignment` (may be "top", "center", or "bottom").
        For compatibility, `ha` and `va` may be used instead.  Either
        a Height `height` or a Width `width` must be specified.  The
        image will be scaled in the unspecified direction to maintain
        the aspect ratio.  If both `height` and `width` are specified,
        then the image's aspect ratio will be ignored.

        Note: As the conversion processes for png and pdf export use
        different libraries, there may be slight differences in output
        depending on output format.

        Optionally, `unitname` may specify a unit for the image,
        i.e. the origin is located at the bototm left corner of the
        image, and (1,1) is located at the upper right corner of the
        image.
        """
        if horizontalalignment is None:
            horizontalalignment = ha if ha is not None else "center"
        if verticalalignment is None:
            verticalalignment = va if va is not None else "center"
        pos_ll = self.convert_to_figure_coord(pos)
        assert height is not None or width is not None, "Either height or width must be given"
        if filename.endswith(".pdf"):
            pdf = mupdf.open(filename)
            page = pdf[0]
            imwidth = page.bound().width
            imheight = page.bound().height
        else:
            with Image.open(filename) as img:
                imwidth = img.size[0]
                imheight = img.size[1]
        if width is None:
            height = self.convert_to_figure_length(height.height())
            width = Width(height.y * (self.size[1]/self.size[0]) * (imwidth/imheight), "figure")
        elif height is None:
            width = self.convert_to_figure_length(width.width())
            height = Height(width.x * (self.size[0]/self.size[1]) * (imheight/imwidth), "figure")
        else:
            height = self.convert_to_figure_length(height.height())
            width = self.convert_to_figure_length(width.width())
        pos_ur = pos_ll + height + width
        if horizontalalignment == "center":
            shift = (pos_ur-pos_ll).width()/2
            pos_ur -= shift
            pos_ll -= shift
        elif horizontalalignment == "right":
            shift = (pos_ur-pos_ll).width()
            pos_ur -= shift
            pos_ll -= shift
        if verticalalignment == "center":
            shift = (pos_ur-pos_ll).height()/2
            pos_ur -= shift
            pos_ll -= shift
        elif verticalalignment == "top":
            shift = (pos_ur-pos_ll).height()
            pos_ur -= shift
            pos_ll -= shift
        assert 0 <= pos_ll.x and 0 <= pos_ll.y and pos_ur.x <= 1 and pos_ur.y <= 1, "Coordinates must not go off the screen"
        filetypes = ["png", "jpg", "jpeg", "gif", "pdf"]
        assert any(ft for ft in filetypes if filename.endswith("."+ft))
        self.images.append((filename, pos_ll, pos_ur))
        if unitname is not None:
            assert self.is_valid_identifier(unitname), f"Invalid axis name {unitname!r}"
            self.add_unit(unitname, (pos_ur-pos_ll), pos_ll)
    def _in_jupyter(self):
        """Test if we are in a Jupyter notebook or in the IPython interpreter."""
        try:
            get_ipython
        except NameError:
            return False
        return True

    def show(self, **kwargs):
        """Display the Canvas in a new window (non-blocking) or in Jupyter.

        Keyword arguments are the same as they are for Canvas.save,
        with the exception of "filename", which is of course not
        available here.
        """
        # Set a low resolution by default so plots aren't too big for the
        # screen
        if "dpi" not in kwargs.keys():
            kwargs['dpi'] = 100
        # Save a temporary image file with the plot
        tmp = tempfile.mkstemp('.png')[1]
        self.tmpfiles.append(tmp)
        self.save(tmp, **kwargs)
        # Display, either in a new window, or in Jupyter
        if self._in_jupyter():
            IPython_display(IPython_Image(filename=tmp))
        else:
            Image.open(tmp).show()
        



# TODO:
# TODO add "thin" to font-manager.py in matplotlib line 81
# TODO set metadata for png/pdf export
# TODO test all fonts for all backends, and figure out pgf backend



# Copyright 2018 Max Shinn <max@maxshinnpotential.com>
# 
# This file is part of Paranoid Scientist, and is available under the
# MIT license.  Please see LICENSE.txt in the root directory for more
# information.

from setuptools import setup

with open("cand/_version.py") as f:
    exec(f.read())

with open("README.md", "r") as f:
    long_desc = f.read()

setup(
    name = 'CanD',
    version = __version__,
    description = 'Create complex layouts for scientific figures in matplotlib',
    long_description = long_desc,
    long_description_content_type='text/markdown',
    author = 'Max Shinn',
    author_email = 'm.shinn@ucl.ac.uk',
    url = 'https://github.com/mwshinn/cand',
    maintainer = 'Max Shinn',
    license = 'MIT',
    python_requires='>=3.5',
    maintainer_email = 'm.shinn@ucl.ac.uk',
    packages = ['cand'],
    install_requires = ['numpy', 'scipy', 'matplotlib', 'paranoid-scientist >= 0.2.1', 'PyMuPDF >= 1.16.0', 'Pillow'],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Framework :: Matplotlib',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Visualization',
    ]
)



import paranoid as pns
import math

class Metric(pns.Type):
    """A Paranoid Scientist Type for Points and Vectors."""
    def test(self, v):
        assert isinstance(v, Point) or isinstance(v, Vector)
    def generate(self):
        yield Point(0, 0)
        yield Point(.3, .1, "absolute")
        yield Point(.1, .1, "figure")
        yield Point(-.1, .1, "figure")
        yield Point(-.1, -3)
        yield Point(2, 2)
        yield Width(.3)
        yield Width(-.1)
        yield Width(0, "figure")
        yield Height(-1, "absolute")
        yield Height(2, "unique")

@pns.paranoidclass
class Point:
    """A point on the canvas in an arbitrary coordinate system.

    A point represents a specific location on the canvas.  It
    represents an x and y coordinate in a specific coordinate system,
    given by `x`, `y`, and `coordinate`, respectively.  Coordinate
    systems are specified as strings, and are interpreted according to
    a Canvas object.

    Points and vectors can be added or subtracted to produce another
    point.  Points cannot be multiplied or divided by scalars.  Points
    cannot be added to each other, but they can be subtracted to
    produce the vector connecting the two points.

    """
    def __new__(cls, x, y, coordinate="default"):
        if isinstance(coordinate, tuple):
            return Point(x, 0, coordinate[0]) >> Point(0, y, coordinate[1])
        else:
            obj = object.__new__(cls)
            obj.x = x
            obj.y = y
            obj.coordinate = coordinate
            return obj
    def __repr__(self):
        nondefault = f', "{self.coordinate}"' if self.coordinate != "default" else ""
        return f'{self.__class__.__name__}({self.x}, {self.y}{nondefault})'
    @pns.accepts(pns.Self, Metric)
    def __add__(self, other):
        """Add together a point and a vector.

        `other` should be a Vector, otherwise this function will throw an error.

        Returns a Point
        """
        if isinstance(other, Vector):
            if self.coordinate == other.coordinate and self.coordinate != "various":
                return Point(self.x + other.x, self.y + other.y, self.coordinate)
            else:
                return BinopPoint(self, '+', other)
        raise ValueError(f"Invalid addition between {self!r} and {other!r}.")
    @pns.accepts(pns.Self, Metric)
    def __sub__(self, other):
        """Find the vector which connects two points.

        `other` should be a Point, otherwise this function will throw an error.

        Returns a Vector.
        """
        if isinstance(other, Vector):
            if self.coordinate == other.coordinate:
                return Point(self.x - other.x, self.y - other.y, self.coordinate)
            else:
                return BinopPoint(self, '-', other)
        elif isinstance(other, Point):
            if self.coordinate == other.coordinate:
                return Vector(self.x - other.x, self.y - other.y, self.coordinate)
            else:
                return BinopVector(self, '-', other)
        raise ValueError(f"Invalid subtraction between {self!r} and {other!r}.")
    def __eq__(self, other):
        """Determine if two Point objects are equal.  

        `other` should be another Point object.

        Returns True or False.
        """
        return (self.x == other.x) and (self.y == other.y) and (self.coordinate == other.coordinate)
    def __iter__(self):
        yield self.x
        yield self.y
    def __rshift__(self, other):
        """Take the x coordinate of this point and the y coordinate of another point.

        `other` should be a Point, otherwise this function will throw an error.

        Returns a Point.
        """
        if not isinstance(other, Point):
            raise ValueError(f"Invalid meet >> operation between {self!r} and {other!r}.")
        if self.coordinate == other.coordinate:
            return Point(self.x, other.y, self.coordinate)
        else:
            return BinopPoint(self, '>>', other)
    def __lshift__(self, other):
        """Take the y coordinate of this point and the x coordinate of another point.

        `other` should be a Point, otherwise this function will throw an error.

        Returns a Point.
        """
        return other >> self
    def __or__(self, other):
        """Return the point in the center of the two given points

        `other` should be a Point, otherwise this function will throw an error.

        Returns a Point.
        """
        if not isinstance(other, Point):
            raise ValueError(f"Invalid mean | operation between {self!r} and {other!r}.")
        if self.coordinate == other.coordinate:
            return Point((self.x+other.x)/2, (self.y+other.y)/2, self.coordinate)
        else:
            return BinopPoint(self, '|', other)

    @staticmethod
    def _generate():
        yield Point(0, 0)
        yield Point(.3, .1, "absolute")
        yield Point(.1, .1, "figure")

@pns.paranoidclass
class Vector:
    """A vector in an arbitrary coordinate system.

    A Vector represents a difference in location on the canvas.  It
    represents a width and a height in a specific coordinate system,
    given by `x`, `y`, and `coordinate`, respectively.  Coordinate
    systems are specified as strings, and are interpreted according to
    a Canvas object.

    Vectors can be added or subtracted to other vectors or points.
    They can also be multiplied and divided by scalars.

    A Vector which only has an x component is automatically cast to a
    "Width" object, and with only a y component is a "Height" object.

    """
    def __new__(cls, x, y, coordinate="default"):
        if isinstance(coordinate, tuple):
            return Vector(x, 0, coordinate[0]) >> Vector(0, y, coordinate[1])
        else:
            obj = object.__new__(cls)
            obj.x = x
            obj.y = y
            obj.coordinate = coordinate
            return obj
    def __repr__(self):
        nondefault = f', "{self.coordinate}"' if self.coordinate != "default" else ""
        return f'{self.__class__.__name__}({self.x}, {self.y}{nondefault})'
    @pns.accepts(pns.Self, Metric)
    def __add__(self, other):
        """Add a vector to a Point or another Vector.

        If `other` is a Point, return a Point.  If `other` is a Vector, return a Vector.
        """
        if isinstance(other, Point):
            return other + self
        elif isinstance(other, Vector):
            if self.coordinate == other.coordinate:
                return Vector(self.x + other.x, self.y + other.y, self.coordinate)
            else:
                return BinopVector(self, '+', other)
        raise ValueError(f"Invalid addition between {repr(self)} and {repr(other)}.")
    @pns.accepts(pns.Self, pns.Self)
    def __sub__(self, other):
        """Vector subtraction.

        `other` should be a Vector.  

        Return a Vector.

        Subtracting a vector is equivalent to adding the negative of the vector.
        """
        if self.coordinate == other.coordinate:
            return Vector(self.x - other.x, self.y - other.y, self.coordinate)
        else:
            return BinopVector(self, '-', other)
    @pns.accepts(pns.Self, pns.Number)
    def __mul__(self, other):
        """Multiply a vector by a scalar.

        `other` should be a scalar by which to multiply each component of the vector.

        Return a Vector.
        """
        return Vector(self.x * other, self.y * other, self.coordinate)
    @pns.accepts(pns.Self, pns.Number)
    def __rmul__(self, other):
        self.__mul__.__doc__
        return self * other
    @pns.accepts(pns.Self)
    def __neg__(self):
        """Take the negative of a vector.

        This is equivalent to rotating the vector by 180 degrees, or
        to multiplying each component by -1.

        Return a Vector.
        """
        return -1* self
    @pns.accepts(pns.Self, pns.Number)
    @pns.requires("other != 0")
    def __truediv__(self, other):
        """Divide a vector by a scalar.

        `other` should be a non-zero scalar by which to divide each component of the vector.

        Return a Vector.
        """
        return Vector(self.x / other, self.y / other, self.coordinate)
    @pns.accepts(pns.Self, pns.Self)
    def __rshift__(self, other):
        """Take the x coordinate of this vector and the y coordinate of another vector.

        `other` should be a Vector, otherwise this function will throw an error.

        Returns a Vector.
        """
        if not isinstance(other, Vector):
            raise ValueError(f"Invalid meet >> operation between {repr(self)} and {repr(other)}.")
        if self.coordinate == other.coordinate:
            return Vector(self.x, other.y, self.coordinate)
        else:
            return BinopVector(self, '>>', other)
    @pns.accepts(pns.Self, pns.Self)
    def __lshift__(self, other):
        """Take the y coordinate of this vector and the x coordinate of another vector.

        `other` should be a Vector, otherwise this function will throw an error.

        Returns a Vector.
        """
        return other >> self
    @pns.accepts(pns.Self, pns.Number)
    def __matmul__(self, other):
        """Rotate the vector by some amount, specified in degrees.

        `other` should be a scalar, in units of degrees.

        Returns a Vector.
        """
        if self.coordinate == "absolute":
            c = math.cos(math.radians(other))
            s = math.sin(math.radians(other))
            return Vector(self.x * c - self.y * s ,
                          self.x * s + self.y * c,
                          self.coordinate)
        else:
            return BinopVector(self, '@', other)
    @pns.accepts(pns.Self, pns.Number)
    def __rmatmul__(self, other):
        self.__matmul__.__doc__
        return self @ other
    @pns.accepts(pns.Self)
    def width(self):
        """Returns a Width object representing the x component of the Vector."""
        return Width(self.x, self.coordinate)
    @pns.accepts(pns.Self)
    def height(self):
        """Returns a Height object representing the y component of the Vector."""
        return Height(self.y, self.coordinate)
    @pns.accepts(pns.Self)
    def flipx(self):
        """Returns a Vector reflected across the y-axis."""
        return Vector(-self.x, self.y, self.coordinate)
    @pns.accepts(pns.Self)
    def flipy(self):
        """Returns a Vector reflected across the x-axis."""
        return Vector(self.x, -self.y, self.coordinate)
    @pns.accepts(pns.Self, pns.Self)
    def __eq__(self, other):
        """Determine if two Vector objects are equal.  

        `other` should be another Vector object.

        Returns True or False.
        """
        return (self.x == other.x) and (self.y == other.y) and (self.coordinate == other.coordinate)
    def __iter__(self):
        yield self.x
        yield self.y
    @classmethod
    def _generate(cls):
        yield Height(.3)
        yield Width(.5)

def Width(x, coordinate="default"):
    """A vector with a 0 in the y coordinate.

    Returns a Vector with `x` in the x coordinate and 0 in the y
    coordinate, within the coordinate system `coordinate`.

    This is included for backward compatibility.
    """
    return Vector(x, 0, coordinate)

def Height(y, coordinate="default"):
    """A vector with a 0 in the x coordinate.

    Returns a Vector with 0 in the x coordinate and `y` in the y
    coordinate, within the coordinate system `coordinate`.

    This is included for backward compatibility.
    """
    return Vector(0, y, coordinate)

@pns.paranoidclass
class MetaBinop:
    coordinate = "various"
    op_table = {'+': lambda lhs,rhs : lhs + rhs,
                '-': lambda lhs,rhs : lhs - rhs,
                '*': lambda lhs,rhs : lhs * rhs,
                '/': lambda lhs,rhs : lhs / rhs,
                '>>': lambda lhs,rhs : lhs >> rhs,
                '|': lambda lhs,rhs : lhs | rhs,
                '@': lambda lhs,rhs : lhs @ rhs}
    @pns.accepts(pns.Unchecked, pns.Or(Metric, pns.Number), pns.Set(['+', '-', '*', '/', '>>', '|', '@']), pns.Or(Metric, pns.Number))
    def __new__(cls, lhs, op, rhs):
        obj = object.__new__(cls)
        obj.lhs = lhs
        obj.op = op
        obj.rhs = rhs
        return obj
    def __repr__(self):
        # Make it look like a MetaBinop with two Points or two Vectors
        # is a point with the tuple-based naming scheme
        if self.op == '>>' and (not isinstance(self.lhs, MetaBinop)) and (not isinstance(self.rhs, MetaBinop)):
            if isinstance(self, BinopPoint):
                classname = "Point"
            elif isinstance(self, BinopVector):
                classname = "Vector"
            return f'{classname}({self.lhs.x}, {self.rhs.y}, ({repr(self.lhs.coordinate)}, {repr(self.rhs.coordinate)}))'
        # Add parens if the point is a binop
        lhstext = f'({repr(self.lhs)})' if isinstance(self.lhs, MetaBinop) and self.lhs.op != '>>' else repr(self.lhs)
        rhstext = f'({repr(self.rhs)})' if isinstance(self.rhs, MetaBinop) and self.rhs.op != '>>' else repr(self.rhs)
        return f'{lhstext} {self.op} {rhstext}'
    def __eq__(self, other):
        return (self.lhs == other.lhs) and (self.op == other.op) and (self.rhs == other.rhs)

@pns.paranoidclass
class BinopPoint(MetaBinop,Point):
    """A Point which is a composite of points and vectors in different bases.

    Each Vector and Point's coordinate system depends on the Canvas
    object, and thus, we cannot immediately compute sums or
    differences of Vector or Point objects.  Instead, this object
    represents a node in a tree of operations on Vector and Point
    objects which results in a Point object.

    """
    @pns.accepts(pns.Self, Vector)
    def __add__(self, rhs):
        if isinstance(rhs, Vector):
            return BinopPoint(self, '+', rhs)
    @pns.accepts(pns.Self, Metric)
    def __sub__(self, rhs):
        if isinstance(rhs, Vector):
            return BinopPoint(self, '-', rhs)
        elif isinstance(rhs, Point):
            return BinopVector(self, '-', rhs)
    def __lshift__(self, rhs):
        return rhs >> self
    def __rshift__(self, rhs):
        if isinstance(rhs, Point):
            return BinopPoint(self, '>>', rhs)
        else:
            raise ValueError(f"Invalid meet >> between {repr(self)} and {repr(other)}.")
    def __or__(self, rhs):
        if isinstance(rhs, Point):
            return BinopPoint(self, '|', rhs)
        else:
            raise ValueError(f"Invalid mean | between {repr(self)} and {repr(other)}.")

@pns.paranoidclass
class BinopVector(MetaBinop,Vector):
    """A Vector which is a composite of points and vectors in different bases.

    Each Vector and Point's coordinate system depends on the Canvas
    object, and thus, we cannot immediately compute sums or
    differences of Vector or Point objects.  Instead, this object
    represents a node in a tree of operations on Vector and Point
    objects which results in a Vector object.

    """
    @pns.accepts(pns.Self, Metric)
    def __add__(self, rhs):
        if isinstance(rhs, Vector):
            return BinopVector(self, '+', rhs)
        elif isinstance(rhs, Point):
            return rhs + self
    @pns.accepts(pns.Self, Vector)
    def __sub__(self, rhs):
        return BinopVector(self, '-', rhs)
    @pns.accepts(pns.Self, pns.Number)
    def __mul__(self, rhs):
        return BinopVector(self, '*', rhs)
    @pns.accepts(pns.Self, pns.Number)
    def __truediv__(self, rhs):
        return BinopVector(self, '/', rhs)
    def __rshift__(self, rhs):
        if isinstance(rhs, Vector):
            return BinopVector(self, '>>', rhs)
        else:
            raise ValueError(f"Invalid meet >> between {repr(self)} and {repr(other)}.")
    def __lshift__(self, rhs):
        return rhs >> self
    @pns.accepts(pns.Self, pns.Number)
    def __matmul__(self, rhs):
        return BinopVector(self, '@', rhs)
    def __rmatmul__(self, rhs):
        rhs @ self
    def width(self):
        """Returns a BinopVector object representing the x component of the Vector."""
        # Handle cases where we use a scalar instead of a vector
        return ((Point(0, 0, "absolute") + self) >> Point(0, 0, "absolute")) - Point(0, 0, "absolute")
    def height(self):
        """Returns a BionpVector object representing the y component of the Vector."""
        return ((Point(0, 0, "absolute") + self) << Point(0, 0, "absolute")) - Point(0, 0, "absolute")
    def flipx(self):
        return (-self) >> self
    def flipy(self):
        return self >> (-self)
    @classmethod
    def _generate(cls):
        yield cls(Point(0, 1), '+', Point(3, 2, "absolute"))
        yield cls(Point(-.2, .2), '-', cls(Point(0, -1, "absolute"), '+', Width(2, "otherunit")))
